#!/usr/bin/python3

#  Copyright (c) 2015 SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# ALL RIGHTS RESERVED.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Neither the name of the SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written
# permission.
#
# This work has been performed in the framework of the SONATA project,
# funded by the European Commission under Grant number 671517 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.sonata-nfv.eu).
#
# This work has also been performed in the framework of the 5GTANGO project,
# funded by the European Commission under Grant number 761493 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.5gtango.eu).

import logging
import coloredlogs
import sys
import os
import yaml
import argparse


log = logging.getLogger(__name__)


class Workspace:
    BACK_CONFIG_VERSION = "0.03"
    CONFIG_VERSION = "0.05"
    DEFAULT_WORKSPACE_DIR = os.path.join(os.path.expanduser("~"),
                                         ".tng-workspace")
    DEFAULT_SCHEMAS_DIR = os.path.join(os.path.expanduser("~"), ".tng-schema")
    __descriptor_name__ = "workspace.yml"

    def __init__(self, ws_root, config=None, ws_name=None, log_level=None):

        self._ws_root = ws_root
        self._ws_config = dict()

        if config:
            self._ws_config = config
        else:
            self.load_default_config()
            if ws_name:
                self.config['name'] = ws_name
            if log_level:
                self.config['log_level'] = log_level

        # coloredlogs.install(level=self.config['log_level'])

    @property
    def workspace_root(self):
        return self._ws_root

    @property
    def workspace_name(self):
        return self.config['name']

    @property
    def default_descriptor_extension(self):
        return self.config['default_descriptor_extension']

    @property
    def log_level(self):
        return self.config['log_level']

    @log_level.setter
    def log_level(self, value):
        self.config['log_level'] = value
        coloredlogs.install(level=value)

    @property
    def schemas_local_master(self):
        return self.config['schemas_local_master']

    @property
    def schemas_remote_master(self):
        return self.config['schemas_remote_master']

    @property
    def validate_watchers(self):
        return self.config['validate_watchers']

    @property
    def config(self):
        return self._ws_config

    @property
    def catalogues_dir(self):
        return self.config['catalogues']

    @property
    def configuration_dir(self):
        return self.config['configuration_dir']

    @property
    def platforms_dir(self):
        return self.config['platforms_dir']

    @property
    def projects_dir(self):
        return self.config['projects_dir']

    @property
    def ns_catalogue_dir(self):
        return os.path.join(self.catalogues_dir, 'ns_catalogue')

    @property
    def vnf_catalogue_dir(self):
        return os.path.join(self.catalogues_dir, 'vnf_catalogue')

    def load_default_config(self):
        self.config['version'] = self.CONFIG_VERSION
        self.config['name'] = '5GTANGO Workspace'
        self.config['log_level'] = 'INFO'

        self.config['catalogues_dir'] = 'catalogues'
        self.config['configuration_dir'] = 'configuration'
        self.config['platforms_dir'] = 'platforms'
        self.config['projects_dir'] = 'projects'
        self.config['projects_config'] = os.path.join(self.workspace_root,
                                                      'projects', 'config.yml')

        self.config['schemas_local_master'] = Workspace.DEFAULT_SCHEMAS_DIR
        self.config['schemas_remote_master'] = \
            "https://github.com/sonata-nfv/tng-schema"

        self.config['default_descriptor_extension'] = 'yml'

        self.config['service_platforms'] = \
            {'sp1':
                {'url': 'http://sp.int3.sonata-nfv.eu:32001',
                 'credentials': {'username': 'sonata',
                                 'password': 's0n@t@',
                                 'token_file': 'token.txt',
                                 'signature': {'pub_key': '',
                                               'prv_key': '',
                                               'cert': ''
                                               }
                                 }
                 }
             }
        self.config['default_service_platform'] = 'sp1'
        self.config['validate_watchers'] = \
            {os.path.join(self.workspace_root, self.config['projects_dir']):
             {'type': 'project',
                'syntax': True,
                'integrity': True,
                'topology': True}
             }

    def create_dirs(self):
        """
        Create the base directory structure for the workspace
        Invoked upon workspace creation.
        :return:
        """
        log.info('Creating workspace at %s', self.workspace_root)
        os.makedirs(self.workspace_root, exist_ok=False)

        dirs = [self.config['catalogues_dir'],
                self.config['configuration_dir'],
                self.config['platforms_dir'],
                self.config['projects_dir'],
                ]

        for d in dirs:
            path = os.path.join(self.workspace_root, d)
            os.makedirs(path, exist_ok=True)

    def create_files(self):
        """
        Creates a workspace configuration file descriptor.
        This is triggered by workspace creation and configuration changes.
        :return:
        """
        cfg_d = self.config.copy()
        ws_file_path = os.path.join(self.workspace_root,
                                    Workspace.__descriptor_name__)

        ws_file = open(ws_file_path, 'w')
        yaml.dump(cfg_d, ws_file, default_flow_style=False)

        # write project config (schema-MIME mapping)
        mapping = {
            'application/vnd.5gtango.vnfd':
                'https://raw.githubusercontent.com/sonata-nfv/tng-schema/master/function-descriptor/vnfd-schema.yml',
            'application/vnd.5gtango.nsd':
                'https://raw.githubusercontent.com/sonata-nfv/tng-schema/master/service-descriptor/nsd-schema.yml',
            'application/vnd.5gtango.tstd':
                'https://raw.githubusercontent.com/sonata-nfv/tng-schema/master/test-descriptor/testdescriptor'
                '-schema.yml',
            'application/vnd.5gtango.slad':
                'https://raw.githubusercontent.com/sonata-nfv/tng-schema/master/sla-template-descriptor/sla-template'
                '-schema.yml',
            'application/vnd.5gtango.rpd':
                'https://raw.githubusercontent.com/sonata-nfv/tng-schema/master/policy-descriptor/policy-schema.yml',
            'application/vnd.5gtango.nstd':
                'https://raw.githubusercontent.com/sonata-nfv/tng-schema/master/slice-descriptor/nst-schema.yml'
        }
        # add reverse mapping (type to schema) for translation of old SONATA descriptors
        reverse_mapping = {v: k for k, v in mapping.items()}
        mapping.update(reverse_mapping)

        conf_path = os.path.join(self.workspace_root, 'projects', 'config.yml')
        conf_file = open(conf_path, 'w')
        yaml.dump(mapping, conf_file, default_flow_style=False)

        return cfg_d

    def check_ws_exists(self):
        ws_file = os.path.join(self.workspace_root,
                               Workspace.__descriptor_name__)
        print(ws_file)
        return os.path.exists(ws_file) or os.path.exists(self.workspace_root)

    @staticmethod
    def load_workspace(ws_root):
        """
        Creates a Workspace object based on a configuration descriptor
        :param ws_root: base path of the workspace
        :return: Workspace object
        """
        if ws_root is None:
            ws_root = Workspace.DEFAULT_WORKSPACE_DIR
        ws_filename = os.path.join(ws_root, Workspace.__descriptor_name__)
        if not os.path.isdir(ws_root) or not os.path.isfile(ws_filename):
            if ws_root == Workspace.DEFAULT_WORKSPACE_DIR:
                # if we tried the default WS, and cannot find it,
                # assume first execution of tng-sdk-project and just
                # create the default WS
                create_workspace(ws_root)
            else:
                log.error("Unable to load workspace descriptor '{}'. "
                          "Create workspace with tng-wks and specify "
                          + "location with -w".format(ws_filename))
                return None

        ws_file = open(ws_filename, 'r')
        try:
            ws_config = yaml.load(ws_file, Loader=yaml.FullLoader)

        except yaml.YAMLError as exc:
            log.error("Error parsing descriptor file '{0}': {1}"
                      .format(ws_filename, exc))
            return
        if not ws_config:
            log.error("Couldn't read descriptor file: '{0}'"
                      .format(ws_filename))
            return

        if not ws_config['version'] == Workspace.CONFIG_VERSION:

            if ws_config['version'] < Workspace.BACK_CONFIG_VERSION:
                log.error("Workspace configuration version '{0}' is no longer "
                          "supported (<{1})"
                          .format(ws_config['version'],
                                  Workspace.BACK_CONFIG_VERSION))
                return
            if ws_config['version'] > \
                    Workspace.CONFIG_VERSION:
                log.error("Workspace configuration version '{0}' is ahead of "
                          "the current supported version (={1})"
                          .format(ws_config['version'],
                                  Workspace.CONFIG_VERSION))
                return

        ws = Workspace(ws_root, config=ws_config)

        # Make adjustments to support backwards compatibility
        # 0.03
        if ws_config['version'] == "0.03":
            ws.config['validate_watchers'] = \
                {os.path.join(ws.workspace_root, ws.config['projects_dir']):
                 {'type': 'project',
                    'syntax': True,
                    'integrity': True,
                    'topology': True}
                 }

        # 0.04
        if ws_config['version'] <= "0.04":
            sps = ws.config['service_platforms']
            for sp_key, sp in sps.items():
                sp['credentials']['signature'] = dict()
                sp['credentials']['signature']['pub_key'] = ''
                sp['credentials']['signature']['prv_key'] = ''
                sp['credentials']['signature']['cert'] = ''

        if ws_config['version'] < Workspace.CONFIG_VERSION:
            log.warning("Loading workspace with an old configuration "
                        "version ({0}). Updated configuration: {1}"
                        .format(ws_config['version'], ws.config))
        return ws

    @property
    def default_service_platform(self):
        return self.config['default_service_platform']

    @default_service_platform.setter
    def default_service_platform(self, sp_id):
        self.config['default_service_platform'] = sp_id

    @property
    def service_platforms(self):
        return self.config['service_platforms']

    @service_platforms.setter
    def service_platforms(self, sps):
        self.config['service_platforms'] = sps

    def get_service_platform(self, sp_id):
        if sp_id not in self.service_platforms.keys():
            return
        return self.service_platforms[sp_id]

    def add_service_platform(self, sp_id):
        if sp_id in self.service_platforms.keys():
            return
        self.service_platforms[sp_id] = {'url': '',
                                         'credentials': {'username': '',
                                                         'password': '',
                                                         'token_file': '',
                                                         'signature': {
                                                             'pub_key': '',
                                                             'prv_key': '',
                                                             'cert': ''
                                                         }}
                                         }

    def config_service_platform(self, sp_id, default=None, url=None,
                                username=None, password=None, token=None,
                                pub_key=None, prv_key=None, cert=None):

        if sp_id not in self.service_platforms.keys():
            return

        sp = self.service_platforms[sp_id]

        if url:
            sp['url'] = url

        if username:
            sp['credentials']['username'] = username

        if password:
            sp['credentials']['password'] = password

        if token:
            sp['credentials']['token_file'] = token

        if pub_key:
            sp['credentials']['signature']['pub_key'] = pub_key

        if prv_key:
            sp['credentials']['signature']['prv_key'] = prv_key

        if cert:
            sp['credentials']['signature']['cert'] = cert

        if default:
            self.default_service_platform = sp_id

        # update workspace config descriptor
        self.create_files()

    def __eq__(self, other):
        """
        Function to assert if two workspaces have the
        same configuration. It overrides the super method
        as is only the need to compare configurations.
        """
        return isinstance(other, type(self)) \
            and self.workspace_name == other.workspace_name \
            and self.workspace_root == other.workspace_root \
            and self.config == other.config


def parse_args_workspace(input_args=None):
    parser = argparse.ArgumentParser(description="Create a new workspace")

    parser.add_argument(
        "-w", "--workspace",
        help="location of new workspace. If not specified will assume '{}'"
              .format(Workspace.DEFAULT_WORKSPACE_DIR),
        required=False)

    parser.add_argument(
        "--debug",
        help="increases logging level to debug",
        required=False,
        action="store_true")

    if input_args is None:
        input_args = sys.argv[1:]
    return parser.parse_args(input_args)


# for entry point tng-workspace; was as "tng-project --init" before
def init_workspace(args=None):
    if args is None:
        args = parse_args_workspace()

    log_level = "INFO"
    if args.debug:
        log_level = "DEBUG"
        coloredlogs.install(level=log_level)

    # If workspace arg is not given, create a workspace in user home
    if not args.workspace:
        ws_root = Workspace.DEFAULT_WORKSPACE_DIR

        # If a workspace already exists at user home, throw an error and quit
        if os.path.isdir(ws_root):
            log.warning("A workspace already exists in {}. Please specify a different location.".format(ws_root))
            return

    else:
        ws_root = os.path.expanduser(args.workspace)
    create_workspace(ws_root, log_level)


def create_workspace(ws_root, log_level="INFO"):
    # create a new workspace
    ws = Workspace(ws_root, log_level=log_level)
    if ws.check_ws_exists():
        log.warning("A workspace already exists at the specified location")
        exit(1)

    log.debug("Attempting to create a new workspace")
    cwd = os.getcwd()
    ws.create_dirs()
    ws.create_files()
    os.chdir(cwd)
    log.debug("Workspace created.")



import os
import oyaml as yaml
import copy


# generate OSM descriptors from the provided high-level arguments
def generate_descriptors(user_inputs, log):
    # load default descriptors (relative to file location, not curr directory)
    log.debug('Loading OSM default descriptors')
    descriptor_dir = os.path.join(os.path.dirname(__file__), os.pardir, 'default-descriptors')
    with open(os.path.join(descriptor_dir, 'osm_default_nsd.yml')) as f:
        osm_default_nsd = yaml.load(f, Loader=yaml.FullLoader)
    with open(os.path.join(descriptor_dir, 'osm_default_vnfd.yml')) as f:
        osm_default_vnfd = yaml.load(f, Loader=yaml.FullLoader)

    # generate VNFDs
    log.debug('Generating 5GTANGO VNFDs')
    vnfds = []
    for i in range(int(user_inputs.vnfs)):
        vnfd = copy.deepcopy(osm_default_vnfd)
        vnfd['vnfd-catalog']['vnfd'][0]['id'] = 'default-vnf{}'.format(i)
        vnfd['vnfd-catalog']['vnfd'][0]['name'] = 'default-vnf{}'.format(i)
        vnfd['vnfd-catalog']['vnfd'][0]['short-name'] = 'default-vnf{}'.format(i)
        vnfd['vnfd-catalog']['vnfd'][0]['vendor'] = user_inputs.vendor

        # add VNF image name and type if available
        if i < len(user_inputs.image_names):
            log.debug("VNF image {}: {}".format(i, user_inputs.image_names[i]))
            vnfd['vnfd-catalog']['vnfd'][0]['vdu'][0]['image'] = \
                user_inputs.image_names[i]
        else:
            log.debug("Using default image for VNF {}".format(i))

        vnfds.append(vnfd)

    # generate NSD
    log.debug('Generating 5GTANGO NSD')
    nsd = osm_default_nsd['nsd-catalog']['nsd'][0]
    nsd['vendor'] = user_inputs.vendor
    nsd['id'] = user_inputs.name
    nsd['name'] = user_inputs.name
    nsd['description'] = user_inputs.description

    # skip first vnf
    for i, vnf in enumerate(vnfds):
        # updated existing entries for vnf0 and then append new ones
        if i > 0:
            nsd['constituent-vnfd'].append({})
            nsd['vld'][0]['vnfd-connection-point-ref'].append({})

        # list involved vnfs
        nsd['constituent-vnfd'][i]['member-vnf-index'] = i
        nsd['constituent-vnfd'][i]['vnfd-id-ref'] = vnf['vnfd-catalog']['vnfd'][0]['id']

        # create mgmt connection points
        nsd['vld'][0]['vnfd-connection-point-ref'][i]['member-vnf-index-ref'] = i
        nsd['vld'][0]['vnfd-connection-point-ref'][i]['vnfd-connection-point-ref'] = 'mgmt'
        nsd['vld'][0]['vnfd-connection-point-ref'][i]['vnfd-id-ref'] = vnf['vnfd-catalog']['vnfd'][0]['id']

    # create vlinks between vnfs
    for i in range(len(vnfds)-1):
        nsd['vld'].append({
            'id': 'vnf{}-2-vnf{}'.format(i, i+1),
            'name': 'vnf{}-2-vnf{}'.format(i, i+1),
            'vnfd-connection-point-ref': [
                {
                    'member-vnf-index-ref': i,
                    'vnfd-connection-point-ref': 'output',
                    'vnfd-id-ref': vnfds[i]['vnfd-catalog']['vnfd'][0]['id']
                },
                {
                    'member-vnf-index-ref': i+1,
                    'vnfd-connection-point-ref': 'input',
                    'vnfd-id-ref': vnfds[i+1]['vnfd-catalog']['vnfd'][0]['id']
                }
            ]
        })

    # create descriptor dictionary
    descriptors = {'nsd': nsd, 'vnfds': vnfds}
    log.info('Generated OSM descriptors at {}'.format(user_inputs.out_path))
    return descriptors



#  Copyright (c) 2015 SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# ALL RIGHTS RESERVED.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Neither the name of the SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written
# permission.
#
# This work has been performed in the framework of the SONATA project,
# funded by the European Commission under Grant number 671517 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.sonata-nfv.eu).
#
# This work has also been performed in the framework of the 5GTANGO project,
# funded by the European Commission under Grant number 761493 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.5gtango.eu).

# this is a namespace package

try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil
    __path__ = pkgutil.extend_path(__path__, __name__)



import os
import oyaml as yaml
import copy


# generate 5GTANGO descriptors from the provided high-level arguments
def generate_descriptors(user_inputs, log):
    # load default descriptors (relative to file location, not curr directory)
    log.debug('Loading 5GTANGO default descriptors')
    descriptor_dir = os.path.join(os.path.dirname(__file__), os.pardir, 'default-descriptors')
    with open(os.path.join(descriptor_dir, 'tango_default_nsd.yml')) as f:
        tango_default_nsd = yaml.load(f, Loader=yaml.FullLoader)
    with open(os.path.join(descriptor_dir, 'tango_default_vnfd.yml')) as f:
        tango_default_vnfd = yaml.load(f, Loader=yaml.FullLoader)

    # generate VNFDs
    log.debug('Generating 5GTANGO VNFDs')
    vnfds = []
    tango_default_vnfd['author'] = user_inputs.author
    tango_default_vnfd['vendor'] = user_inputs.vendor
    for i in range(int(user_inputs.vnfs)):
        vnfd = copy.deepcopy(tango_default_vnfd)
        vnfd['name'] = 'default-vnf{}'.format(i)

        # add VNF image name if available
        if i < len(user_inputs.image_names):
            log.debug("VNF {} image name: {}"
                      .format(i, user_inputs.image_names[i]))
            vnfd['virtual_deployment_units'][0]['vm_image'] = \
                user_inputs.image_names[i]
        else:
            log.debug("Using default image name for VNF {}".format(i))
        # add VNF image name if available
        if i < len(user_inputs.image_types):
            log.debug("VNF {} image type: {}"
                      .format(i, user_inputs.image_types[i]))
            vnfd['virtual_deployment_units'][0]['vm_image_format'] = \
                user_inputs.image_types[i]
        else:
            log.debug("Using default image type for VNF {}".format(i))

        vnfds.append(vnfd)

    # generate NSD
    log.debug('Generating 5GTANGO NSD')
    nsd = tango_default_nsd
    nsd['author'] = user_inputs.author
    nsd['vendor'] = user_inputs.vendor
    nsd['name'] = user_inputs.name
    nsd['description'] = user_inputs.description

    # keep reusable parts of the default NSD that are independent from #VNFs
    # eg, there is always a mgmt vLink and a vLink from the input for vnf0
    # remove other stuff which will be replaced (vnf0-2-output vLink)
    del nsd['virtual_links'][2]
    del nsd['forwarding_graphs'][0]['constituent_virtual_links'][1]
    del nsd['forwarding_graphs'][0]['network_forwarding_paths'][0]['connection_points'][2:4]

    # now updated and extend the NSD
    for i, vnf in enumerate(vnfds):
        # list of involved VNFs
        # first entry already exists -> adjust, then append new ones
        if i > 0:
            nsd['network_functions'].append({})
        nsd['network_functions'][i]['vnf_id'] = 'vnf{}'.format(i)
        nsd['network_functions'][i]['vnf_name'] = vnf['name']
        nsd['network_functions'][i]['vnf_vendor'] = vnf['vendor']
        nsd['network_functions'][i]['vnf_version'] = vnf['version']

        # create corresponding vLinks
        # add vLink to next vnf
        if i < len(vnfds) - 1:
            nsd['virtual_links'].append({
                'id': 'vnf{}-2-vnf{}'.format(i, i+1),
                'connectivity_type': 'E-Line',
                'connection_points_reference': [
                    'vnf{}:output'.format(i), 'vnf{}:input'.format(i+1)
                ]
            })
        # for last vnf in chain, set vLink to output instead
        else:
            nsd['virtual_links'].append({
                'id': 'vnf{}-2-output'.format(i, i+1),
                'connectivity_type': 'E-Line',
                'connection_points_reference': [
                    'vnf{}:output'.format(i), 'output'
                ]
            })

        # add mgmt connection point (already exists for vnf0)
        if i > 0:
            nsd['virtual_links'][0]['connection_points_reference'].append(
                'vnf{}:mgmt'.format(i)
            )

    # adjust forwarding graph
    nsd['forwarding_graphs'][0]['number_of_virtual_links'] = len(vnfds) + 1
    # append new vLinks (skip mgmt and input-2-vnf0, which are already there)
    for i in range(2, len(nsd['virtual_links'])):
        nsd['forwarding_graphs'][0]['constituent_virtual_links'].append(
            nsd['virtual_links'][i]['id']
        )
    # append new vnfs
    for i in range(1, len(vnfds)):
        nsd['forwarding_graphs'][0]['constituent_vnfs'].append(
            nsd['network_functions'][i]['vnf_id']
        )

    # append in- and output of each vLink (keep input, vnf0:input, vnf0:output)
    pos = 3
    for i in range(2, len(nsd['virtual_links'])):
        nsd['forwarding_graphs'][0]['network_forwarding_paths'][0]['connection_points'].append({
            'connection_point_ref': nsd['virtual_links'][i]['connection_points_reference'][0],
            'position': pos
        })
        pos += 1
        nsd['forwarding_graphs'][0]['network_forwarding_paths'][0]['connection_points'].append({
            'connection_point_ref': nsd['virtual_links'][i]['connection_points_reference'][1],
            'position': pos
        })
        pos += 1

    # create descriptor dictionary
    descriptors = {'nsd': nsd, 'vnfds': vnfds}
    log.info('Generated 5GTANGO descriptors at {}'.format(user_inputs.out_path))
    return descriptors



#!/usr/bin/python3

#  Copyright (c) 2015 SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# ALL RIGHTS RESERVED.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Neither the name of the SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written
# permission.
#
# This work has been performed in the framework of the SONATA project,
# funded by the European Commission under Grant number 671517 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.sonata-nfv.eu).
#
# This work has also been performed in the framework of the 5GTANGO project,
# funded by the European Commission under Grant number 761493 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.5gtango.eu).

import os
import logging
import oyaml as yaml        # ordered yaml to avoid reordering of descriptors
import glob
import mimetypes
from collections import defaultdict
from tabulate import tabulate
from tngsdk.project.workspace import Workspace
from tngsdk.descriptorgen import descriptorgen

log = logging.getLogger(__name__)


class Project:
    CONFIG_VERSION = "0.5"

    __descriptor_name__ = 'project.yml'

    def __init__(self, workspace, prj_root, config=None):
        # be able to hanlde different workspace inputs
        if workspace is None or isinstance(workspace, str):
            # workspace is a string
            # treat it as path and auto-load
            workspace = Workspace.load_workspace(workspace)
        self._prj_root = prj_root
        self._workspace = workspace
        self.error_msg = None

        if config:
            self._prj_config = config
        else:
            self.load_default_config()

        # get config from workspace for URL->MIME mapping
        with open(workspace.config["projects_config"], 'r') as config_file:
            self.type_mapping = yaml.load(config_file, Loader=yaml.FullLoader)

    @property
    def project_root(self):
        return self._prj_root

    @property
    def nsd_root(self):
        return os.path.join(self._prj_root, 'sources', 'nsd')

    @property
    def vnfd_root(self):
        return os.path.join(self._prj_root, 'sources', 'vnf')

    @property
    def project_config(self):
        return self._prj_config

    @property
    def descriptor_extension(self):
        return self.project_config['descriptor_extension']

    def load_default_config(self):
        self._prj_config = {
            'version': self.CONFIG_VERSION,
            'package': {
                'name': '5gtango-project-sample',
                'vendor': 'eu.5gtango',
                'version': '0.1',
                'maintainer': 'Name, Company, Contact',
                'description': 'Some description about this sample'
            },
            'descriptor_extension':
                self._workspace.default_descriptor_extension,
            'files': []
        }

    # create new project (empty or with descriptors by descriptorgen)
    def create_prj(self, args):
        # create project root directory (if it doesn't exist)
        log.info('Creating project at {}'.format(self._prj_root))
        if os.path.isdir(self._prj_root):
            log.error("Unable to create project at '{}'. Directory already exists.".format(self._prj_root))
            exit(1)
        os.makedirs(self._prj_root, exist_ok=False)

        # create subdirs, sample descriptors, and project.yml
        if args.empty:
            log.debug('Creating empty project (no folders or sample files)')
        elif args.vnfs <= 0:
            log.info('Num. VNFs set to 0 (or below). Creating empty project.')
        else:
            self._gen_descriptors(args)
        self._write_prj_yml()

    # generate descriptors using the descriptorgen module and specified args
    def _gen_descriptors(self, args):
        args.out_path = self._prj_root
        log.info("Generating descriptors")
        log.debug("Descriptor generation args: {}".format(args))
        descriptorgen.generate(args)

        # add generated files to project manifest
        log.debug("Adding generated descriptors to project manifest")
        self.add_file(os.path.join(self._prj_root, "*"))

    # writes project descriptor to file (project.yml)
    def _write_prj_yml(self):
        prj_path = os.path.join(self._prj_root, Project.__descriptor_name__)
        with open(prj_path, 'w') as prj_file:
            prj_file.write(yaml.dump(self._prj_config,
                                     default_flow_style=False))

    # resolves wildcards by calling add/remove_file for each file
    def resolve_wildcards(self, path, add=False, remove=False):
        for f in glob.glob(path):
            if add:
                self.add_file(f)
            if remove:
                self.remove_file(f)

    # detects and returns MIME type of specified file
    def mime_type(self, file):
        # tag folders to be zipped by packager
        if os.path.isdir(file):
            type = 'application/vnd.folder.compressed.zip'

        elif os.path.isfile(file):
            name, extension = os.path.splitext(file)

            # check yml files to detect and classify 5GTANGO descriptors
            if extension == ".yml" or extension == ".yaml":
                with open(file, 'r') as yml_file:
                    yml_file = yaml.load(yml_file, Loader=yaml.FullLoader)
                    if 'descriptor_schema' in yml_file:
                        type = self.type_mapping[yml_file['descriptor_schema']]
                    # try to detect OSM descriptors based on field names
                    elif 'constituent-vnfd' in yml_file and 'vld' in yml_file:
                        type = 'application/vnd.etsi.osm.nsd'
                    elif 'vnfd-catalog' in yml_file:
                        type = 'application/vnd.etsi.osm.vnfd'
                    else:
                        log.warning('Could not detect MIME type of {}. '
                                    'Using text/yaml'.format(file))
                        type = 'text/yaml'

            # for non-yml files determine the type using mimetypes
            else:
                (type, encoding) = mimetypes.guess_type(file, strict=False)
                # add more types from a config with mimetypes.read_mime_types(file)

        else:
            log.warning('{} is not a file or directory. Not added to project.'.format(file))
            return -1

        log.debug('Detected MIME type: {}'.format(type))
        return type

    # adds a file to the project: detects type and adds to project.yml
    def add_file(self, file_path, type=None):
        # resolve wildcards
        if '*' in file_path:
            log.debug('Attempting to resolve wildcard in {}'.format(file_path))
            self.resolve_wildcards(file_path, add=True)
            return

        # try to detect the MIME type if none is given
        if type is None:
            type = self.mime_type(file_path)
            # path is not a file or directory -> ignore and don't add
            if type == -1:
                return
        if type is None:
            log.warning('Could not detect MIME type of {}. Using "application/octet-stream".'.format(file_path))
            self.error_msg = 'Could not detect MIME type of {}. Using "application/octet-stream".'.format(file_path)
            type = "application/octet-stream"       # default MIME type

        # set tags accordingly
        tags = []
        if '5gtango' in type:
            tags = ['eu.5gtango']
        elif 'osm' in type:
            tags = ['etsi.osm']

        # calculate relative file path to project root
        abs_file_path = os.path.abspath(file_path)
        abs_prj_root = os.path.abspath(self._prj_root)
        rel_file_path = os.path.relpath(abs_file_path, abs_prj_root)
        # fix windows paths by replacing \ with /
        if os.name == 'nt':
            rel_file_path = rel_file_path.replace('\\', '/')
            log.debug('Adjusted Windows path in project.yml: {}'.format(rel_file_path))

        # add to project.yml
        file = {'path': rel_file_path, 'type': type, 'tags': tags}
        if file in self._prj_config['files']:
            log.warning('{} is already in project.yml.'.format(file_path))
        else:
            self._prj_config['files'].append(file)
            self._write_prj_yml()
            log.info('Added {} to project.yml'.format(file_path))

    # removes a file from the project
    def remove_file(self, file_path):
        # resolve wildcards
        if '*' in file_path:
            log.debug('Attempting to resolve wildcard in {}'.format(file_path))
            self.resolve_wildcards(file_path, remove=True)
            return

        # calculate relative file path to project root (similar to add_file)
        abs_file_path = os.path.abspath(file_path)
        abs_prj_root = os.path.abspath(self._prj_root)
        rel_file_path = os.path.relpath(abs_file_path, abs_prj_root)
        # adjust to windows paths by replacing \ with /
        if os.name == 'nt':
            rel_file_path = rel_file_path.replace('\\', '/')
            log.debug('Adjusted relative Windows path to match project.yml: {}'.format(rel_file_path))

        for f in self._prj_config['files']:
            if f['path'] == rel_file_path:
                self._prj_config['files'].remove(f)
                self._write_prj_yml()
                log.info('Removed {} from project.yml'.format(file_path))
                return
        log.warning('{} is not in project.yml'.format(file_path))

    # prints project info/status
    def status(self):
        # print general info
        print('Project: {}'.format(self._prj_config['package']['name']))
        print('Vendor: {}'.format(self._prj_config['package']['vendor']))
        print('Version: {}'.format(self._prj_config['package']['version']))
        print(self._prj_config['package']['description'])

        if 'files' not in self._prj_config:
            log.warning('Old SONATA project: project.yml does not have a files section!'
                        'To translate an old SONATA project to a new 5GTANGO project, use --translate')
            return

        # collect and print info about involved MIME types (type + quanity)
        types = defaultdict(int)
        for f in self._prj_config['files']:
            types[f['type']] += 1
        print(tabulate(types.items(), headers=['MIME type', 'Quantity'], tablefmt='grid'))

    # translate old SONATA VNFD or NSD to new 5GTANGO format (descriptor_version --> descriptor_schema)
    def translate_descriptor(self, descriptor_file, vnfd):
        log.info('Translating descriptor {}'.format(descriptor_file))
        with open(descriptor_file, 'r') as f:
            descriptor = yaml.load(f, Loader=yaml.FullLoader)

        descriptor.pop('descriptor_version')
        if vnfd:
            schema = self.type_mapping['application/vnd.5gtango.vnfd']
        else:
            schema = self.type_mapping['application/vnd.5gtango.nsd']
        descriptor['descriptor_schema'] = schema

        with open(descriptor_file, 'w') as f:
            f.write(yaml.dump(descriptor, default_flow_style=False))

    # translate old SONATA project to new 5GTANGO project (in place)
    def translate(self):
        log.debug('Attempting to translate old SONATA project to new 5GTANGO '
                  'project (in place): {}'.format(self._prj_root))

        # update/set version number to current version
        log.debug('Updating version number to {}'.format(self.CONFIG_VERSION))
        self._prj_config['version'] = self.CONFIG_VERSION

        # update descriptors: replace "schema_version" with "descriptor_schema" based on nsd/vnf folder
        log.debug('Updating old SONATA descriptors to new 5GTANGO descriptors')
        vnfd_path = os.path.join(self._prj_root, 'sources', 'vnf', '**', '*.yml')
        for vnfd in glob.glob(vnfd_path, recursive=True):
            if os.path.isfile(vnfd):
                self.translate_descriptor(vnfd, vnfd=True)
        nsd_path = os.path.join(self._prj_root, 'sources', 'nsd', '**', '*.yml')
        for nsd in glob.glob(nsd_path, recursive=True):
            if os.path.isfile(nsd):
                self.translate_descriptor(nsd, vnfd=False)

        # create files section and add files
        log.debug('Creating "files" section and adding all files in {}'.format(self._prj_root))
        self._prj_config['files'] = []
        for f in glob.glob(os.path.join(self._prj_root, 'sources', '**'), recursive=True):
            if os.path.isfile(f):
                self.add_file(f)

        self._write_prj_yml()
        log.info('Successfully translated {} to 5GTANGO project.'.format(self._prj_root))

    # return a list of relative file paths to all NSDs (default: Tango NSDs)
    def get_nsds(self, type='application/vnd.5gtango.nsd'):
        return self.get_file_paths(type)

    # return a list of relative file paths to all VNFDs (default: Tango VNFDs)
    def get_vnfds(self, type='application/vnd.5gtango.vnfd'):
        return self.get_file_paths(type)

    # return a list of relative file paths to all TSTDs (default: Tango TSTDs)
    def get_tstds(self, type='application/vnd.5gtango.tstd'):
        return self.get_file_paths(type)

    def get_slads(self, type='application/vnd.5gtango.slad'):
        return self.get_file_paths(type)

    def get_rpds(self, type='application/vnd.5gtango.rpd'):
        return self.get_file_paths(type)

    def get_nstds(self, type='application/vnd.5gtango.nstd'):
        return self.get_file_paths(type)

    # return a list of relative (to proj root) file paths to files of the specified type
    def get_file_paths(self, type):
        return [f['path'] for f in self._prj_config['files'] if f['type'] == type]

    @staticmethod
    def __is_valid__(project):
        """Checks if a given project is valid"""
        if type(project) is not Project:
            return False

        if not os.path.isfile(os.path.join(
                project.project_root,
                Project.__descriptor_name__)):
            return False

        return True

    # loads a project using its project manifest (project.yml)
    @staticmethod
    def load_project(prj_root, workspace=None, translate=False):
        # load default workspace if none specified
        if workspace is None:
            workspace = Workspace.load_workspace(Workspace.DEFAULT_WORKSPACE_DIR)

        # check if project manifest exists
        prj_filename = os.path.join(prj_root, Project.__descriptor_name__)
        if not os.path.isdir(prj_root) or not os.path.isfile(prj_filename):
            log.error("Unable to load project manifest '{}'".format(prj_filename))
            return None

        # load project manifest
        log.info("Loading project '{}'".format(prj_filename))
        with open(prj_filename, 'r') as prj_file:
            try:
                prj_config = yaml.load(prj_file, Loader=yaml.FullLoader)
            except yaml.YAMLError as exc:
                log.error("Error parsing descriptor file: {0}".format(exc))
                return
            if not prj_config:
                log.error("Couldn't read descriptor file: '{0}'".format(prj_file))
                return

        # create a new project object with the same manifest
        if prj_config['version'] == Project.CONFIG_VERSION:
            return Project(workspace, prj_root, config=prj_config)

        # deal with different versions
        if prj_config['version'] < Project.CONFIG_VERSION and not translate:
            log.warning("Project version {} is outdated (current: {}). To translate to new 5GTANGO project version use "
                        "--translate".format(prj_config['version'], Project.CONFIG_VERSION))
        if prj_config['version'] > Project.CONFIG_VERSION:
            log.warning("Project version {} is ahead of the current version {}."
                        .format(prj_config['version'], Project.CONFIG_VERSION))

        return Project(workspace, prj_root, config=prj_config)



#!/usr/bin/python3

#  Copyright (c) 2015 SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# ALL RIGHTS RESERVED.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Neither the name of the SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written
# permission.
#
# This work has been performed in the framework of the SONATA project,
# funded by the European Commission under Grant number 671517 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.sonata-nfv.eu).
#
# This work has also been performed in the framework of the 5GTANGO project,
# funded by the European Commission under Grant number 761493 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.5gtango.eu).

import logging
import argparse
import sys
import os
from tngsdk.project.workspace import Workspace
from tngsdk.project.project import Project

log = logging.getLogger(__name__)


def parse_args(input_args=None):
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description="5GTANGO SDK project")

    parser.add_argument("-v", "--debug",
                        help="increases logging level to debug",
                        required=False,
                        action="store_true")

    # project CLI
    parser.add_argument("-p", "--project",
                        help="create a new project at the specified location",
                        required=False,
                        default='new-project')

    parser.add_argument("-w", "--workspace",
                        help="location of existing (or new) workspace. "
                        "If not specified will assume '{}'"
                        .format(Workspace.DEFAULT_WORKSPACE_DIR),
                        required=False)

    parser.add_argument("--empty",
                        help="create an empty project (without sample files)",
                        required=False,
                        action="store_true")

    parser.add_argument("--add",
                        help="Add file to project",
                        required=False,
                        default=None)

    parser.add_argument("-t", "--type",
                        help="MIME type of added file (only with --add)",
                        required=False,
                        default=None)

    parser.add_argument("--remove",
                        help="Remove file from project",
                        required=False,
                        default=None)

    parser.add_argument("--status",
                        help="Show project file paths",
                        required=False,
                        action="store_true")

    parser.add_argument("--translate",
                        help="Translate old SONATA project to new 5GTANGO project",
                        required=False,
                        action="store_true")

    # descriptorgen CLI
    parser.add_argument('-o', help='set relative output path',
                        required=False, default='.', dest='out_path')
    parser.add_argument('--tango', help='only generate 5GTANGO descriptors',
                        required=False, action='store_true')
    parser.add_argument('--osm', help='only generate OSM descriptors',
                        required=False, action='store_true')
    parser.add_argument('--author', help='set a specific NSD and VNFD author',
                        required=False, default='5GTANGO Developer', dest='author')
    parser.add_argument('--vendor', help='set a specific NSD and VNFD vendor',
                        required=False, default='eu.5gtango', dest='vendor')
    parser.add_argument('--name', help='set a specific NSD name',
                        required=False, default='tango-nsd', dest='name')
    parser.add_argument('--description', help='set a specific NSD description',
                        required=False, default='Default description',
                        dest='description')
    parser.add_argument('--vnfs', help='set a specific number of VNFs',
                        type=int, required=False, default=1, dest='vnfs')
    parser.add_argument('--image_names',
                        help='list of VNF image names (default: ubuntu)',
                        nargs='*', required=False, default='')
    parser.add_argument('--image_types',
                        help='list of VNF image types (default: docker)',
                        nargs='*', required=False, default='')

    # service management
    parser.add_argument("-s", "--service",
                        help="Run tng-project in service mode with REST API.",
                        required=False,
                        default=False,
                        dest="service",
                        action="store_true")

    parser.add_argument("--dump-swagger",
                        help="Dump Swagger JSON of REST API and exit",
                        required=False,
                        default=False,
                        dest="dump_swagger",
                        action="store_true")

    parser.add_argument("--address",
                        help="Listen address of REST API when in service mode.",
                        required=False,
                        default="0.0.0.0",
                        dest="service_address")

    parser.add_argument("--port",
                        help="TCP port of REST API when in service mode.",
                        required=False,
                        default=5098,
                        dest="service_port")

    if input_args is None:
        input_args = sys.argv[1:]
    return parser.parse_args(input_args)


# handle cli input to create/modify project
def dispatch(args):
    if args is None:
        args = parse_args()

    # use specified workspace or default
    if args.workspace:
        ws_root = os.path.expanduser(args.workspace)
    else:
        ws_root = Workspace.DEFAULT_WORKSPACE_DIR

    ws = Workspace.load_workspace(ws_root)
    if not ws:
        log.error("Could not find a 5GTANGO workspace at the specified location")
        exit(1)

    prj_root = os.path.expanduser(args.project)

    if args.add:
        # load project and add file to project.yml
        log.debug("Attempting to add file {}".format(args.add))
        proj = Project.load_project(prj_root, ws)
        proj.add_file(args.add, type=args.type)

    elif args.remove:
        # load project and remove file from project.yml
        log.debug("Attempting to remove file {}".format(args.remove))
        proj = Project.load_project(prj_root, ws)
        proj.remove_file(args.remove)

    elif args.status:
        # load project and show status
        log.debug("Attempting to show project status")
        proj = Project.load_project(prj_root, ws)
        proj.status()

    elif args.translate:
        proj = Project.load_project(prj_root, ws, translate=True)
        proj.translate()

    else:
        # create project
        log.debug("Attempting to create a new project")

        if args.vnfs != len(args.image_names):
            log.info("Number of VNFs and VNF image names don't match."
                     " Using default image names if necessary.")
        if args.vnfs != len(args.image_types):
            log.info("Number of VNFs and VNF image types don't match."
                     " Using default image types if necessary.")

        proj = Project(ws, prj_root)
        proj.create_prj(args)
        log.debug("Project created.")

    return proj



#  Copyright (c) 2015 SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# ALL RIGHTS RESERVED.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Neither the name of the SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written
# permission.
#
# This work has been performed in the framework of the SONATA project,
# funded by the European Commission under Grant number 671517 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.sonata-nfv.eu).
#
# This work has also been performed in the framework of the 5GTANGO project,
# funded by the European Commission under Grant number 761493 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.5gtango.eu).

import logging
import subprocess
import json
import os
import uuid
import shutil
import zipfile
from flask import Flask, Blueprint, send_from_directory
from flask_restplus import Resource, Api, Namespace, fields, inputs
from werkzeug.middleware.proxy_fix import ProxyFix
from werkzeug.datastructures import FileStorage
from tngsdk import cli
# important: import as cli_project; else would collide with Project class here
from tngsdk.project.project import Project as cli_project
from flask_cors import CORS
from os import listdir
from os.path import isfile, join
import ntpath

log = logging.getLogger(__name__)

app = Flask(__name__)
CORS(app)
app.wsgi_app = ProxyFix(app.wsgi_app)
blueprint = Blueprint('api', __name__, url_prefix="/api")
api_v1 = Namespace("v1", description="tng-project API v1")
api = Api(blueprint, version="0.1", title='5GTANGO tng-project API',
          description="5GTANGO tng-project REST API for project management.")
app.register_blueprint(blueprint)
api.add_namespace(api_v1)


# parser arguments: for input parameters sent to the API
project_parser = api_v1.parser()
project_parser.add_argument("author",
                            required=False,
                            default='eu.tango',
                            help="Service author")
project_parser.add_argument("vendor",
                            required=False,
                            default='eu.tango',
                            help="Service vendor")
project_parser.add_argument("name",
                            required=False,
                            default='example-service',
                            help="Service name")
project_parser.add_argument("description",
                            required=False,
                            default='Example description',
                            help="Service description")
project_parser.add_argument("vnfs",
                            type=int,
                            required=False,
                            default=1,
                            help="Number of VNFs in the service")
project_parser.add_argument("image_names",
                            type=str,
                            required=False,
                            help="List of VNF image names (space-separated)")
project_parser.add_argument("image_types",
                            type=str,
                            required=False,
                            help="List of VNF image types (space-separated)")
project_parser.add_argument("only_tango",
                            # do not use "bool"!
                            # https://github.com/noirbizarre/flask-restplus/issues/199#issuecomment-276645303
                            type=inputs.boolean,
                            required=False,
                            default=False,
                            help="Generate only 5GTANGO descriptors")
project_parser.add_argument("only_osm",
                            type=inputs.boolean,
                            required=False,
                            default=False,
                            help="Generate only OSM descriptors")

file_upload_parser = api_v1.parser()
file_upload_parser.add_argument("file",
                                location="files",
                                type=FileStorage,
                                required=True,
                                help="Uploaded file to add to project")
file_upload_parser.add_argument("file_type",
                                required=False,
                                default=None,
                                help="MIME type of an uploaded file")

filename_parser = api_v1.parser()
filename_parser.add_argument("filename", required=True, help="Filename of the file to remove")

package_parser = api_v1.parser()
package_parser.add_argument("skip_validation", required=False, type=inputs.boolean, default=False,
                            help="If true, skip validation when packaging. Else validate first.")

# models for marshaling return values from the API (also used for generating Swagger spec)
ping_get_model = api_v1.model("PingGet", {
    "alive_since": fields.String(description="system uptime", required=True)
})
projects_get_model = api_v1.model("ProjectsGet", {
    "projects": fields.List(fields.String, description="list of all project UUIDs", required=True)
})
projects_post_model = api_v1.model("ProjectsPost", {
    "uuid": fields.String(description="project UUID", required=True),
    "error_msg": fields.String(description="error message"),
    "files": fields.List(fields.String, description="list of files along with access points", required=True)
})
project_get_model = api_v1.model("ProjectGet", {
    "project_uuid": fields.String(description="project UUID", required=True),
    "manifest": fields.String(description="project manifest", required=True),
    "error_msg": fields.String(description="error message", required=True),
})
project_delete_model = api_v1.model("ProjectDelete", {
    "project_uuid": fields.String(description="project UUID", required=True)
})
# model per file used for nested files type
file_model = api_v1.model("File", {
    'path': fields.String(attribute='path'),
    'type': fields.String(attribute='type'),
    'tags': fields.List(fields.String)
})
files_get_model = api_v1.model("FilesGet", {
    "project_uuid": fields.String(description="project UUID"),
    "files": fields.List(fields.Nested(api_v1.models['File']), description="list of all project files"),
    "error_msg": fields.String(description="error message")
})
files_post_model = api_v1.model("FilesPost", {
    "project_uuid": fields.String(description="project UUID"),
    "filename": fields.String(description="added file"),
    "error_msg": fields.String(description="error message")
})
files_delete_model = api_v1.model("FilesDelete", {
    "project_uuid": fields.String(description="project UUID"),
    "removed_file": fields.String(description="deleted file"),
    "error_msg": fields.String(description="error message")
})

package_post_model = api_v1.model("PackagePost", {
    "project_uuid": fields.String(description="Project UUID"),
    "package_name": fields.String(description="Name of the created package"),
    "package_path": fields.String(description="Path of the created package"),
    "error_msg": fields.String(description="Error message")
})


def dump_swagger():
    """Dump Swagger specification in docs/rest_api.json"""
    # TODO replace this with the URL of a real tng-project service
    app.config.update(SERVER_NAME="tng-project.5gtango.eu")
    with app.app_context():
        with open(os.path.join("docs", "rest_api.json"), "w") as f:
            f.write(json.dumps(api.__schema__))


# start REST API server
def serve_forever(args, debug=True):
    """Start serving (forever) the REST API with flask"""
    # TODO replace this with WSGIServer for better performance
    log.info("Starting tng-project in service mode")
    app.cliargs = args
    app.run(host=args.service_address, port=args.service_port, debug=debug)


@api_v1.route("/pings")
class Ping(Resource):
    @api_v1.marshal_with(ping_get_model)
    def get(self):
        """Health check: Respond with current uptime"""
        uptime = None
        try:
            uptime = str(subprocess.check_output("uptime")).strip()
        except BaseException as e:
            log.warning(str(e))
        return {"alive_since": uptime}


@api_v1.route("/projects")
class Projects(Resource):
    @api_v1.marshal_with(projects_get_model)
    def get(self):
        """Get list of projects"""
        log.info("GET to /projects. Loading available projects")
        os.makedirs('projects', exist_ok=True)
        project_dirs = [name for name in os.listdir('projects') if os.path.isdir(os.path.join('projects', name))]
        return {'projects': project_dirs}

    @api_v1.expect(project_parser)
    @api_v1.marshal_with(projects_post_model)
    def post(self):
        """Create a new project and generate descriptors with the given args"""
        args = project_parser.parse_args()
        log.info("POST to /projects with args: {}".format(args))

        # transform args into array passed to the descriptorgen cli
        # create a new UUID for each project (to avoid whitespaces)
        project_uuid = str(uuid.uuid4())
        dgn_args = ['-p', os.path.join('projects', project_uuid)]

        for k, v in args.items():
            # only add --tango/--osm if value=True
            if k == 'only_tango':
                if v:
                    dgn_args.append('--tango')
            elif k == 'only_osm':
                if v:
                    dgn_args.append('--osm')
            elif k == 'image_names' or k == 'image_types':
                if v is not None:
                    dgn_args.append('--' + k)
                    # split multiple image names/types and append all of them
                    dgn_args.extend(v.split(' '))
            else:
                # convert #vnfs from int to str (CLI only accepts string)
                if k == 'vnfs':
                    v = str(v)
                # add '--' as prefix for each key to be consistent with CLI
                dgn_args.append('--' + k)
                dgn_args.append(v)
        log.debug("CLI args: {}".format(dgn_args))

        cli_args = cli.parse_args(dgn_args)
        project = cli.dispatch(cli_args)

        project_path = os.path.join('projects', project_uuid)
        project_files = [f for f in listdir(project_path) if isfile(join(project_path, f))]
        info = {'uuid': project_uuid, "error_msg": project.error_msg, "files": []}
        files = []
        for f in project_files:
            files.append('http://127.0.0.1:5098/api/v1/projects/' + project_uuid + '/' + f)

        info['files'] = files
        return info


@api_v1.route("/projects/<string:project_uuid>")
class Project(Resource):
    @api_v1.marshal_with(project_get_model)
    @api_v1.response(200, 'OK')
    @api_v1.response(404, "Project not found")
    def get(self, project_uuid):
        """Get the UUID and manifest of the specified project"""
        log.info("GET to /projects/{}".format(project_uuid))
        project_path = os.path.join('projects', project_uuid)
        if not os.path.isdir(project_path):
            log.error("No project found with name/UUID {}".format(project_uuid))
            return {'error_msg': "Project not found: {}".format(project_uuid)}, 404

        project = cli_project.load_project(project_path)
        return {"project_uuid": project_uuid, "manifest": project.project_config, "error_msg": project.error_msg}

    @api_v1.marshal_with(project_delete_model)
    @api_v1.response(200, 'OK')
    @api_v1.response(404, "Project not found")
    def delete(self, project_uuid):
        """Delete the specified project"""
        log.info("DELETE to /projects/{}".format(project_uuid))
        project_path = os.path.join('projects', project_uuid)
        if not os.path.isdir(project_path):
            log.error("No project found with name/UUID {}".format(project_uuid))
            return {'error_msg': "Project not found: {}".format(project_uuid)}, 404

        shutil.rmtree(project_path)
        return {"project_uuid": project_uuid}


@api_v1.route("/projects/<string:project_uuid>/<string:file_name>")
class ProjectSpecificFile(Resource):
    def get(self, project_uuid, file_name):
        """Get/download the specified file from the specified project"""
        log.info("GET to /projects/{}/{}".format(project_uuid, file_name))
        projects_dir = os.path.realpath('projects')
        project_path = os.path.join(projects_dir, project_uuid)
        log.debug("Project path: {}".format(project_path))
        return send_from_directory(project_path, file_name)


@api_v1.route("/projects/<string:project_uuid>/files")
class ProjectFiles(Resource):
    # get list of project files
    @api_v1.marshal_with(files_get_model)
    @api_v1.response(200, 'OK')
    @api_v1.response(404, "Project not found")
    def get(self, project_uuid):
        """Get a list of files in the specified project"""
        log.info("GET to /projects/{}/files".format(project_uuid))
        project_path = os.path.join('projects', project_uuid)
        if not os.path.isdir(project_path):
            log.error("No project found with name/UUID {}".format(project_uuid))
            return {'error_msg': "Project not found: {}".format(project_uuid)}, 404

        project = cli_project.load_project(project_path)
        return {"project_uuid": project_uuid, "files": project.project_config["files"]}

    @api_v1.expect(file_upload_parser)
    @api_v1.marshal_with(files_post_model)
    @api_v1.response(200, 'OK')
    @api_v1.response(404, "Project not found")
    def post(self, project_uuid):
        """Upload and add a file to the specified project"""
        args = file_upload_parser.parse_args()
        log.info("POST to /projects/{}/files with args: {}".format(project_uuid, args))

        # try to load the project
        project_path = os.path.join('projects', project_uuid)
        if not os.path.isdir(project_path):
            log.error("No project found with name/UUID {}".format(project_uuid))
            return {'error_msg': "Project not found: {}".format(project_uuid)}, 404
        project = cli_project.load_project(project_path)

        # check if file already exists
        file = args["file"]
        if os.path.isfile(os.path.join(project_path, file.filename)):
            log.warning("Overriding existing file {}".format(file.filename))

        # save uploaded file to project and add to project manifest
        log.debug("Adding uploaded file {} to project with UUID {}".format(file.filename, project_uuid))
        file.save(os.path.join(project_path, file.filename))
        project.add_file(os.path.join(project_path, file.filename), args["file_type"])

        return {"project_uuid": project_uuid, "filename": file.filename, "error_msg": project.error_msg}

    @api_v1.expect(filename_parser)
    @api_v1.marshal_with(files_delete_model)
    @api_v1.response(200, 'OK')
    @api_v1.response(404, "Project or file not found")
    def delete(self, project_uuid):
        """Delete the specified project"""
        args = filename_parser.parse_args()
        log.info("DELETE to /projects/{}/files with args: {}".format(project_uuid, args))

        # try to load the project
        project_path = os.path.join('projects', project_uuid)
        if not os.path.isdir(project_path):
            log.error("No project found with name/UUID {}".format(project_uuid))
            return {'error_msg': "Project not found: {}".format(project_uuid)}, 404
        project = cli_project.load_project(project_path)

        # check if file exists
        filename = args["filename"]
        if not os.path.isfile(os.path.join(project_path, filename)):
            log.error("File {} not found in project with UUID {}".format(filename, project_uuid))
            return {"project_uuid": project_uuid, "error_msg": "File {} not found in project".format(filename)}, 404

        # remove file from project manifest and delete it
        project.remove_file(os.path.join(project_path, filename))
        os.remove(os.path.join(project_path, filename))
        return {"project_uuid": project_uuid, "removed_file": filename, "error_msg": project.error_msg}


# not needed and implementation not completed; just left for possible future use
@api_v1.deprecated
@api_v1.route("/projects/<string:project_uuid>/download")
class ProjectDownload(Resource):
    """Download the specified project as zip file. Not completely implemented. Use packager instead."""
    @api_v1.response(200, 'OK')
    @api_v1.response(404, "Project not found")
    def get(self, project_uuid):
        log.info("GET to /projects/{}/download".format(project_uuid))

        # try to load the project
        project_path = os.path.join('projects', project_uuid)
        if not os.path.isdir(project_path):
            log.error("No project found with name/UUID {}".format(project_uuid))
            return {'error_msg': "Project not found: {}".format(project_uuid)}, 404
        # project = cli_project.load_project(project_path)

        # zip the project
        zip_path = os.path.join('projects', project_uuid + '.zip')
        zipped_project = zipfile.ZipFile(zip_path, 'w')
        for root, dirs, files in os.walk(project_path):
            for f in files:
                zipped_project.write(os.path.join(root, f), os.path.relpath(os.path.join(root, f), 'projects'))
        zipped_project.close()

        # TODO: return the zipped project; async?
        return "not implemented", 501


@api_v1.route("/projects/<string:project_uuid>/package")
class ProjectPackage(Resource):
    @api_v1.expect(package_parser)
    @api_v1.marshal_with(package_post_model)
    @api_v1.response(200, 'OK')
    @api_v1.response(404, "Project not found")
    @api_v1.response(503, "tng-sdk-package not installed")
    def post(self, project_uuid):
        """Package (and validate) the specified project using tng-sdk-package (if installed)"""
        args = package_parser.parse_args()
        log.info("POST to /projects/{}/package with args: {}".format(project_uuid, args))

        # try to load the project
        project_path = os.path.join('projects', project_uuid)
        if not os.path.isdir(project_path):
            log.error("No project found with name/UUID {}".format(project_uuid))
            return {'error_msg': "Project not found: {}".format(project_uuid)}, 404

        # try to import the packager if it's installed
        try:
            import tngsdk.package
        except BaseException as ex:
            log.error("Cancel packaging: tng-sdk-package not installed")
            log.debug(ex)
            return {'error_msg': "Cancel packaging: tng-sdk-package not installed"}, 503

        # call the packager
        pkg_args = [
            '--package', project_path,
            '--output', project_path,
        ]
        if args['skip_validation']:
            log.debug("Skipping validation")
            pkg_args.append('--skip-validation')
        r = tngsdk.package.run(pkg_args)
        log.debug(r)
        if r.error is not None:
            return {'project_uuid': project_uuid, 'package_name': None, 'package_path': None,
                    'error_msg': "Package error: {}".format(r.error)}
        pkg_path = r.metadata.get("_storage_location")
        pkg_name = ntpath.basename(pkg_path)
        log.debug("Package name {} and path {}".format(pkg_name, pkg_path))
        return {'project_uuid': project_uuid, 'package_name': pkg_name, 'package_path': pkg_path, 'error_msg': None}



#  Copyright (c) 2015 SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# ALL RIGHTS RESERVED.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Neither the name of the SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written
# permission.
#
# This work has been performed in the framework of the SONATA project,
# funded by the European Commission under Grant number 671517 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.sonata-nfv.eu).
#
# This work has also been performed in the framework of the 5GTANGO project,
# funded by the European Commission under Grant number 761493 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.5gtango.eu).

import os
from setuptools import setup, find_packages


cwd = os.path.dirname(__file__)
with open(os.path.join(cwd, 'requirements.txt')) as f:
    requirements = f.read().splitlines()

setup(name='tngsdk.project',
      license='Apache License, Version 2.0',
      version='5.1.0',
      url='https://github.com/sonata-nfv/tng-sdk-project',
      author='Manuel Peuster, Stefan Schneider',
      author_email='manuel.peuster@upb.de, stefan.schneider@upb.de',
      package_dir={'': 'src'},
      packages=find_packages('src'),   # dependency resolution
      include_package_data=True,       # package data specified in MANIFEST.in
      install_requires=requirements,
      zip_safe=False,
      entry_points={
          'console_scripts': [
              'tng-sdk-workspace=tngsdk.project.workspace:init_workspace',
              'tng-workspace=tngsdk.project.workspace:init_workspace',
              'tng-wks=tngsdk.project.workspace:init_workspace',
              'tng-sdk-project=tngsdk.project:main',
              'tng-project=tngsdk.project:main',
              'tng-prj=tngsdk.project:main',
              'tng-sdk-descriptorgen=tngsdk.descriptorgen.descriptorgen:'
              'generate',
              'tng-descriptorgen=tngsdk.descriptorgen.descriptorgen:generate',
              'tng-dgn=tngsdk.descriptorgen.descriptorgen:generate'
          ],
      },
      test_suite='tngsdk',
      tests_require=['pytest', 'tavern', 'flake8'])



#!/usr/bin/python3

#  Copyright (c) 2015 SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# ALL RIGHTS RESERVED.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Neither the name of the SONATA-NFV, 5GTANGO, UBIWHERE, Paderborn University
# nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written
# permission.
#
# This work has been performed in the framework of the SONATA project,
# funded by the European Commission under Grant number 671517 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.sonata-nfv.eu).
#
# This work has also been performed in the framework of the 5GTANGO project,
# funded by the European Commission under Grant number 761493 through
# the Horizon 2020 and 5G-PPP programmes. The authors would like to
# acknowledge the contributions of their colleagues of the SONATA
# partner consortium (www.5gtango.eu).


import oyaml as yaml        # ordered yaml to avoid reordering of descriptors
import os
import logging
import coloredlogs
from tngsdk import cli
from tngsdk.descriptorgen.plugins import tango, osm

log = logging.getLogger(__name__)


# save the generated descriptors in the specified folder; add a prefix for each flavor
def save_descriptors(nsd, vnfds, flavor, folder='.'):
    # create dir if it doesn't exist
    if not os.path.exists(folder):
        os.makedirs(folder)

    # dump generated nsd and vnfds
    outfile = os.path.join(folder, '{}_nsd.yml'.format(flavor))
    with open(outfile, 'w', newline='') as f:
        yaml.dump(nsd, f, default_flow_style=False)
    for i, vnf in enumerate(vnfds):
        outfile = os.path.join(folder, '{}_vnfd{}.yml'.format(flavor, i))
        with open(outfile, 'w', newline='') as f:
            yaml.dump(vnf, f, default_flow_style=False)


def generate(args=None):
    if args is None:
        args = cli.parse_args()

    if args.debug:
        coloredlogs.install(level='DEBUG')
    else:
        coloredlogs.install(level='INFO')

    log.info("Generating descriptors with args {}".format(args))

    # generate and save tango descriptors
    if not args.osm:
        descriptors = tango.generate_descriptors(args, log)
        save_descriptors(descriptors['nsd'], descriptors['vnfds'], 'tango', args.out_path)

    # generate and save osm descriptors
    if not args.tango:
        descriptors = osm.generate_descriptors(args, log)
        save_descriptors(descriptors['nsd'], descriptors['vnfds'], 'osm', args.out_path)


if __name__ == '__main__':
    generate()
